/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

static char rcsid[] = "$Id: dyna.c,v 1.6 1997/01/22 15:19:00 apg Exp $";

#include "tac_plus.h"

#ifdef DYNA
#include "dynipd.h"

#define LOCALHOST       "127.0.0.1"
#define UNKN            "SERVER-FAIL"

static  u_long  saddr  = 0;
static  int     serial = 0;
static  u_long  lport  = 0;
static  char    db_lastip[MAXARGBUF]   = UNKN;
char            db_lastname[MAXARGBUF] = "";

/*
 * Resolve and store dynamic IP host name. Also optional port
 * number can be supplied after hostname using colon, e.g.
 * mv.demos.su:20005
 */
void    dyna_sethost(hname)
	char    *hname;
{
	struct  hostent         *hp;
	char                    *p;

	serial = getpid();

	if (p = strchr(hname, ':')) {
		*p = '\0';
		if (sscanf(++p, "%d", &lport) != 1) {
			report(LOG_ERR, "bad dynipd port: %s\n", p);
			tac_exit(1);
		}
	}
	if (!(saddr = inet_addr(hname))) {
		if (!(hp = gethostbyname(hname))) {
			report(LOG_ERR, ": can't resolve dynamic ip server address (%s)\n", hname);
			tac_exit(1);
		}
		saddr = *((u_long *) hp->h_addr);
	}
	report(LOG_DEBUG, "dynip server addr set to %s, lport now %u\n",
		inet_ntoa(*((struct in_addr *) &saddr)), lport);
}

#define RFAIL   if (sd > 0) (void) close(sd); return tac_strdup(UNKN);

/*
 * Process server request
 * (also serve duplicate requests - cache last successful one)
 */
char   *dyna_req(name, fix)
	char    *name;
	int     fix;
{
	struct sockaddr_in      sca;
	struct  dynp            dd;
	int                     sd = -1, one = 1;
	char                    *pool = NULL;

	if (!strcmp(db_lastname, name))
		return (tac_strdup(db_lastip));

	pool  = cfg_get_pvalue(name, TAC_IS_USER,
	    S_apool, TAC_PLUS_RECURSE);
#ifdef UDB
	if (pool && *pool == '%') {
		DB      *db = (DB *) cfg_get_udb(name);
		if (!db) {
			report(LOG_ERR, "No database assigned for user %s\n", name);
			RFAIL
		}
		pool = (char *) db_replace(name, db, pool);
		if (*pool == '%') {
			report(LOG_ERR, ": can't subst apool %s from db\n", pool);
			RFAIL
		}
	}
#endif /* UDB */
	if (!saddr)
		saddr = inet_addr(LOCALHOST);

	if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		report(LOG_ERR, ": can't get socket, errno %d\n", errno);
		RFAIL
	}
	if (setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE,
	    (char *) &one, sizeof (one)) < 0) {
		report(LOG_ERR, "setsockopt KEEPALIVE, errno %d", errno);
		RFAIL
	}
	one = sizeof (dd);
	if (setsockopt(sd, SOL_SOCKET, SO_SNDLOWAT,
	    (char *) &one, sizeof (one)) < 0) {
		report(LOG_ERR, "setsockopt SNDLOWAT, errno %d", errno);
		RFAIL
	}
	bzero((char *) &sca, sizeof (sca));
	sca.sin_family      = AF_INET;
	sca.sin_addr.s_addr = saddr;
	sca.sin_port        = htons(lport ? lport : DPORT);

	if (connect(sd, (struct sockaddr *) &sca, sizeof(sca)) < 0) {
		report(LOG_ERR, ": can't connect to dynip server, addr %s\n",
			inet_ntoa(saddr));
		RFAIL
	}
	dd.serial = serial;
	dd.magic  = DYNP_MAGIC;
	if (pool) {
		dd.op     = DYNP_IARQ2;
		(void) strncpy(&dd.argb[MAXARGBUF/2], pool, MAXARGBUF/2 - 1);
	} else
		dd.op     = DYNP_IARQ;

	(void) strncpy(dd.argb, name, MAXARGBUF/2 - 1);

	if (write(sd, (char *) &dd, 1) != 1 ||
	    write(sd, (char *) &dd, sizeof(dd)) != sizeof(dd)) {
		report(LOG_ERR, ": can't write to dynip socket\n");
		RFAIL
	}
	dd.serial = dd.op = dd.magic = 0;
	(void) read(sd, (char *) &dd, 1);
	if (read(sd, (char *) &dd, sizeof(dd)) != sizeof(dd)) {
		report(LOG_ERR, ": reading dynip socket\n");
		RFAIL
	}
	if (dd.serial != serial || dd.magic != DYNP_MAGIC ||
	    (dd.op != DYNP_IAREPL && dd.op != DYNP_IERROR)) {
		report(LOG_ERR, ": protocol error, bad serial/magic/op\n");
		RFAIL
	}
	if (dd.op == DYNP_IERROR) {
		report(LOG_ERR, ": dynip server error: %s\n", dd.argb);
		RFAIL
	}
	if (!inet_addr(dd.argb)) {
		report(LOG_ERR, ": dynip server returned bad ipn (%s)\n", dd.argb);
		RFAIL
	}
	(void) close(sd);

	(void) strcpy(db_lastname, name);          /* cache last request */
	(void) strcpy(db_lastip,   dd.argb);

	return tac_strdup(dd.argb);
}
#endif /* DYNA */
