
#include "tac_plus.h"

#include <signal.h>
#include <pwd.h>
#include <stdarg.h>

#ifndef TAC_RESOLV_FILE
 #define TAC_RESOLV_FILE TAC_PLUS_CFGFILE
#endif

#define tac_abort { if(f >= 0) close(f); return session.aborted = 1; }

struct session session;
static int inited = 0;

static void catchup(s)
{
  printf("\n*** ERROR: TACACS+ SERVER NOT RESPONDING!\n");
  session.aborted = 1;
  return;
}

int tac_connect()
{

  int f = -1;
  FILE *cf;
  struct sockaddr_in s;
  void (*oldalrm)();
  char buf[256];

  if(!inited)
   {
    bzero(&session, sizeof(session));
    inited++;
   }

  if(session.peer)
   {
    free(session.peer);
    session.peer = NULL;
   }
  if(session.key)
   {
    free(session.key);
    session.key = NULL;
   }
  session.aborted = 0;

  if ((cf = fopen(TAC_RESOLV_FILE, "r")) == NULL) return 1;
  while((session.peer == NULL || session.key == NULL) &&
        fgets(buf, sizeof(buf), cf)) {
    char *s, *v;
    s = strtok(buf, " \t\n");
    v = strtok(NULL, " \t\n");
    if (session.peer == NULL && !strcmp(s, "server"))
     {
       session.peer = tac_strdup(v);
       continue;
     }
    if (session.key == NULL && !strcmp(s, "key"))
        session.key = tac_strdup(v);
  }
  fclose(cf);

  if(session.peer == NULL || session.key == NULL) tac_abort;

  if ((f = socket(AF_INET, SOCK_STREAM, 0)) < 0) tac_abort;
  s.sin_addr.s_addr = htonl(INADDR_ANY);
#ifdef HAVE_SIN_LEN
  s.sin_len = sizeof(struct sockaddr_in);
#endif
  s.sin_family = AF_INET;
  s.sin_port = 0;
  if (bind(f, (struct sockaddr *)&s, sizeof(s)) < 0) tac_abort;
  if (!inet_aton(session.peer, &s.sin_addr.s_addr)) tac_abort;
#ifdef HAVE_SIN_LEN
  s.sin_len = sizeof(struct sockaddr_in);
#endif
  s.sin_family = AF_INET;
  s.sin_port = htons(49);

  oldalrm = signal(SIGALRM, catchup);
  alarm(TAC_PLUS_CONNECT_TIMEOUT);
  if (connect(f, (struct sockaddr *)&s, sizeof(s)) < 0) tac_abort;
  alarm(0);
  session.sock = f;

  session.session_id = htonl(getpid());
/*
  session.session_id = getpid();
*/
  session.seq_no = 0;
  session.last_exch = time(NULL);
  signal(SIGALRM, oldalrm);
  return session.aborted;
}

int send_auth_start(l)
char *l;
{
  size_t len;
  char buf[256];
  HDR *hdr = (HDR *)buf;
  struct authen_start *ask = (struct authen_start *)(buf + TAC_PLUS_HDR_SIZE);
  char *p = buf + TAC_PLUS_HDR_SIZE + TAC_AUTHEN_START_FIXED_FIELDS_SIZE;

  bzero(buf, sizeof(buf));

  strncpy(p, l, sizeof(buf) - (TAC_PLUS_HDR_SIZE + TAC_AUTHEN_START_FIXED_FIELDS_SIZE + 1));
  len = strlen(p);

  hdr->version = TAC_PLUS_VER_0;
  hdr->type = TAC_PLUS_AUTHEN;
  hdr->seq_no = ++session.seq_no;
  hdr->encryption = TAC_PLUS_CLEAR;
  hdr->session_id = session.session_id;
  hdr->datalength = htonl(TAC_AUTHEN_START_FIXED_FIELDS_SIZE+len);

  ask->action = TAC_PLUS_AUTHEN_LOGIN;
  ask->priv_lvl = TAC_PLUS_PRIV_LVL_MIN;
  ask->authen_type = TAC_PLUS_AUTHEN_TYPE_ASCII;
  ask->service = TAC_PLUS_AUTHEN_SVC_LOGIN;
  ask->port_len = len;

  if (write_packet(buf)) return 1;

  return 0;
}

int send_auth_cont(up)
char *up;
{
  size_t len;
  char buf[256];
  HDR *hdr = (HDR *)buf;
  struct authen_cont *ask = (struct authen_cont *)(buf + TAC_PLUS_HDR_SIZE);
  char *p = (char *)
    (buf + TAC_PLUS_HDR_SIZE + TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE);

  bzero(buf, sizeof(buf));

  strncpy(p, up, sizeof(buf) - (TAC_PLUS_HDR_SIZE + TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE + 1));
  len = strlen(p);

  hdr->version = TAC_PLUS_VER_0;
  hdr->type = TAC_PLUS_AUTHEN;
  hdr->seq_no = ++session.seq_no;
  hdr->encryption = TAC_PLUS_CLEAR;
  hdr->session_id = session.session_id;
  hdr->datalength = htonl(len + TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE);

  ask->user_msg_len = htons(len);

  if (write_packet(buf)) return 1;

  return 0;
}

extern char *read_packet();

char *tac_ident(u, p, l)
char *u, *p, *l;
{
  char *tmp;
  int mlen, dlen;
  char *mptr, *dptr;
  struct authen_reply *rep;
  char *pw;

  if (send_auth_start(l)) return 0;
  if (tmp = read_packet()) free(tmp); else return 0;
  if (send_auth_cont(u)) return 0;;
  if (tmp = read_packet()) free(tmp); else return 0;
  if (send_auth_cont(p)) return 0;
  tmp = read_packet();
  if (!tmp) return 0;
  /*  . .. */
  rep = (struct authen_reply *)(tmp + TAC_PLUS_HDR_SIZE);

  mlen = ntohs(rep->msg_len);
  mptr = (char *)(tmp+TAC_PLUS_HDR_SIZE+TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE);
  if (mlen) { write(1, mptr, mlen); write(1, "\n\n", 2); }
  if (rep->status != TAC_PLUS_AUTHEN_STATUS_PASS) { free(tmp); return 0; }
  dlen = ntohs(rep->data_len);
  dptr = mptr + mlen;
  if (!dlen) { free(tmp); return 0; }
  pw = tac_strdup(dptr);
  free(tmp);
  return pw;
}

struct passwd *tacacs_plus_auth(user, password, port)
char *user, *password, *port;
{
  char *s, *p;
  static struct passwd pw;

  if (tac_connect()) {
    perror("*** ERROR: UNABLE TO AUTHENTICATE");
    return(NULL);
  }
  s = tac_ident(user, password, port);
  close(session.sock);
  if (!s) return 0;
  p = strtok(s, ":");
  if (pw.pw_name) free(pw.pw_name);
  pw.pw_name = tac_strdup(p);
  p = strtok(NULL, ":");
  pw.pw_uid = atoi(p);
  p = strtok(NULL, ":");
  pw.pw_gid = atoi(p);
  p = strtok(NULL, ":");
  if (pw.pw_dir) free(pw.pw_dir);
  pw.pw_dir = tac_strdup(p ? p : "");
  p = strtok(NULL, ":");
  if (pw.pw_shell) free(pw.pw_shell);
  pw.pw_shell = tac_strdup(p ? p : "");
  free(s);
  return &pw;
}

/*
 * stst:
 * TAC_PLUS_ACCT_FLAG_MORE
 * TAC_PLUS_ACCT_FLAG_START
 * TAC_PLUS_ACCT_FLAG_STOP
 */
int tacacs_plus_acct(char *user, char *line, int stst, ...)
{
  va_list ap;
  char buf[4096];
  HDR *hdr = (HDR *)buf;
  struct acct *acc = (struct acct *)(buf + TAC_PLUS_HDR_SIZE);
  int len = 0, c;
  char *p;
  int ac = 0;
  u_char *av;
  size_t ul = strlen(user),
         ll = strlen(line),
         hl;

  if (tac_connect()) {
    perror("*** ERROR: UNABLE TO AUTHENTICATE");
    return(NULL);
  }

  va_start(ap, stst);
  for (ac=0; ; ac++) {
    p = va_arg(ap, char *);
    if(!p) break;
    len += strlen(p);
  }

  hl = strlen(session.peer);

  len += ac + ul + ll + hl;

  if(len + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE >= sizeof(buf)) {
    perror("*** ERROR: TOO LONG ARGUMENT\n*** ERROR: UNABLE TO AUTHENTICATE");
    return(NULL);
  }

  bzero(buf, sizeof(buf));

  hdr->version = TAC_PLUS_VER_0;
  hdr->type = TAC_PLUS_ACCT;
  hdr->seq_no = ++session.seq_no;
  hdr->encryption = TAC_PLUS_CLEAR;
  hdr->session_id = session.session_id;
  hdr->datalength = htonl(len+TAC_ACCT_REQ_FIXED_FIELDS_SIZE);

  acc->flags = stst;
  acc->authen_method = AUTHEN_METH_TACACSPLUS;
  acc->priv_lvl = TAC_PLUS_PRIV_LVL_MIN;
  acc->authen_type = TAC_PLUS_AUTHEN_TYPE_ASCII;
  acc->authen_service = TAC_PLUS_AUTHEN_SVC_LOGIN;
  acc->user_len = ul;
  acc->port_len = ll;
  acc->rem_addr_len = hl;
  acc->arg_cnt = ac;

  av = (char *)(buf + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE);
  p = av + ac;
  strcpy(p, user);
  p += ul;
  strcpy(p, line);
  p += ll;
  strcpy(p, session.peer);
  p += hl;
  va_start(ap, stst);
  for (c=0; c<ac; c++) {
    strcpy(p, va_arg(ap, char *));
    p += av[c] = strlen(p);
  }

  if (write_packet(buf)) return 1;
  if (p = read_packet()) free(p); else return 1;
  return 0;
}

int debug = 0;
int console = 0;
int single = 0;


/* DSS */

int send_auth_start_pap(user, pwd, port)
char *user, *pwd, *port;
{
  size_t len, user_len, pwd_len, port_len;
  char buf[256];
  HDR *hdr = (HDR *)buf;
  struct authen_start *ask = (struct authen_start *)(buf + TAC_PLUS_HDR_SIZE);
  char *p = buf + TAC_PLUS_HDR_SIZE + TAC_AUTHEN_START_FIXED_FIELDS_SIZE;

  bzero(buf, sizeof(buf));

  user_len = strlen(user); if(user_len > 16) user_len = 16;
  strncpy(p, user, user_len); p = p + user_len;
  port_len = strlen(port); if(port_len > 16) port_len = 16;
  strncpy(p, port, port_len); p = p + port_len;
  pwd_len = strlen(pwd); if(pwd_len > 16) pwd_len = 16;
  strncpy(p, pwd, pwd_len); p = p + pwd_len;
  len = user_len + port_len + pwd_len;

  hdr->version = TAC_PLUS_VER_1;
  hdr->type = TAC_PLUS_AUTHEN;
  hdr->seq_no = ++session.seq_no;
  hdr->encryption = TAC_PLUS_CLEAR;
  hdr->session_id = session.session_id;
  hdr->datalength = htonl(TAC_AUTHEN_START_FIXED_FIELDS_SIZE + len);

  ask->action = TAC_PLUS_AUTHEN_LOGIN;
  ask->priv_lvl = TAC_PLUS_PRIV_LVL_MIN;
  ask->authen_type = TAC_PLUS_AUTHEN_TYPE_PAP;
  ask->service = TAC_PLUS_AUTHEN_SVC_PPP;
  ask->user_len = user_len;
  ask->port_len = port_len;
  ask->data_len = pwd_len;

  if (write_packet(buf)) return 1;

  return 0;
}

int tacacs_plus_auth_pap(user, pwd, port)
char *user, *pwd, *port;
{
  char *tmp;
  struct authen_reply *rep;

  if (tac_connect()) {
    perror("*** ERROR: UNABLE TO AUTHENTICATE");
    return 0;
  }

  if(send_auth_start_pap(user, pwd, port)) return 0;
  tmp = read_packet();
  if(!tmp) return 0;
  /*  . .. */
  rep = (struct authen_reply *)(tmp + TAC_PLUS_HDR_SIZE);
  if (rep->status != TAC_PLUS_AUTHEN_STATUS_PASS) { free(tmp); return 0; }
  free(tmp); return 1;
}
